@@ -1,16 +1,18 @@
 # -*- coding: UTF-8 -*-
 import lief
 import math
-import matplotlib.pyplot as plt
 import os
+import re
 import statistics
+from functools import lru_cache
 
 
 __all__ = ["bintropy", "characteristics", "entropy", "is_packed", "plot", "THRESHOLDS"]
 
 
+__btype = lambda b: str(type(b)).split(".")[2]
 __log = lambda l, m, lvl="debug": getattr(l, lvl)(m) if l else None
-__secname = lambda s: s.strip("\x00") or s or"<empty>"
+__secname = lambda s: s.strip("\x00") or s or "<empty>"
 
 # https://matplotlib.org/2.0.2/examples/color/named_colors.html
 COLORS = {
@@ -46,7 +48,7 @@ COLORS = {
     'literal4': "mediumblue",     # 8-byte literal values
     'common':   "royalblue",      # uninitialized imported symbol definitions
 }
-MIN_ZONE_WIDTH = 5  # minimum number of samples on the entropy plot for a section (so that it can still be visible even
+MIN_ZONE_WIDTH = 3  # minimum number of samples on the entropy plot for a section (so that it can still be visible even
                     #  if it is far smaller than the other sections)
 N_SAMPLES = 2048
 SUBLABELS = {
@@ -67,7 +69,25 @@ THRESHOLDS = {
     #TODO: get average and highest entropy values for lief.EXE_FORMATS.MACHO
 }
 
-plt.rcParams['font.family'] = "serif"
+
+def _get_ep_and_section(binary):
+    """ Helper for computing the entry point and finding its section for each supported format.
+    :param binary: LIEF-parsed binary object
+    :return:       (ep_file_offset, name_of_ep_section)
+    """
+    btype = __btype(binary)
+    try:
+        if btype in ["ELF", "MachO"]:
+            ep = binary.virtual_address_to_offset(binary.entrypoint)
+            ep_section = binary.section_from_offset(ep)
+        elif btype == "PE":
+            ep = binary.rva_to_offset(binary.optional_header.addressof_entrypoint)
+            ep_section = binary.section_from_rva(binary.optional_header.addressof_entrypoint)
+        else:
+            raise OSError("Unknown format")
+        return ep, ep_section.name
+    except (AttributeError, TypeError):
+        return None, None
 
 
 def _human_readable_size(size, precision=0):
@@ -79,8 +99,24 @@ def _human_readable_size(size, precision=0):
     return "%.*f%s" % (precision, size, units[i])
 
 
+# dirty fix for section names requiring to get their real names from the string table (as lief does not seem to handle
+#  this in every case)
+@lru_cache
+def _real_section_names(path):
+    from subprocess import check_output
+    try:
+        names, out = [], check_output(["objdump", "-h", path]).decode("latin-1")
+    except FileNotFoundError:
+        return
+    for l in out.split("\n"):
+        m = re.match(r"\s+\d+\s(.*?)\s+", l)
+        if m:
+            names.append(m.group(1))
+    return names
+
+
 def bintropy(executable, mode=0, blocksize=256, ignore_half_block_zeros=True, decide=True,
-             threshold_average_entropy=None, threshold_highest_entropy=None, logger=None, **kwargs):
+             threshold_average_entropy=None, threshold_highest_entropy=None, logger=None, parsed=None, **kwargs):
     """ Simple implementation of Bintropy as of https://ieeexplore.ieee.org/document/4140989.
     
     :param executable:                path to the executable to be analyzed
@@ -91,15 +127,13 @@ def bintropy(executable, mode=0, blocksize=256, ignore_half_block_zeros=True, de
     :param threshold_average_entropy: threshold on average entropy for deciding if packed
     :param threshold_highest_entropy: threshold on highest entropy for deciding if packed
     :param logger:                    logger instance for debug purpose
+    :param parsed:                    already parsed binary object
     :return:                          if decide is True  => bool (whether the input executable is packed or not)
                                                    False => (average_entropy, highest_block_entropy)
     """
-    # try to parse the binary first ; capture the stderr messages from LIEF
-    tmp_fd, null_fd = os.dup(2), os.open(os.devnull, os.O_RDWR)
-    os.dup2(null_fd, 2)
-    binary = lief.parse(str(executable))
-    os.dup2(tmp_fd, 2)  # restore stderr
-    os.close(null_fd)
+    path = str(executable)
+    # try to parse the binary first
+    binary = parsed or lief.parse(path)
     if binary is None:
         raise OSError("Unknown format")
     # now select the right thresholds
@@ -108,7 +142,7 @@ def bintropy(executable, mode=0, blocksize=256, ignore_half_block_zeros=True, de
     _t1, _t2 = [_t1, thresholds[0]][_t1 is None], [_t2, thresholds[1]][_t2 is None]
     # FIRST MODE: compute the entropy of the whole executable
     if mode == 0:
-        with open(str(executable), 'rb') as f:
+        with open(path, 'rb') as f:
             exe = f.read()
         __log(logger, "Entropy (Shannon): {}".format(entropy(exe)))
         e = entropy(exe, blocksize, ignore_half_block_zeros)
@@ -135,8 +169,13 @@ def bintropy(executable, mode=0, blocksize=256, ignore_half_block_zeros=True, de
         e, w = {}, {}
         if mode == 1:  # per section
             if len(binary.sections) > 0:
-                for section in binary.sections:
-                    n, d = section.name, section.content
+                for i, section in enumerate(binary.sections):
+                    n, d = section.name, section.content.tobytes()
+                    # special case: for some executables compiled with debug information, some sections may be named
+                    #                with the format "/[N]" where N is the offset in the string table ; in this case, we
+                    #                compute the real section names
+                    if re.match(r"^\/\d+$", n) and _real_section_names(path):
+                        n = _real_section_names(path)[i]
                     _handle(n, d)
             else:  # in some cases, packed executables can have no section ; e.g. UPX(/bin/ls)
                 __log(logger, "This file has no section", "error")
@@ -144,7 +183,7 @@ def bintropy(executable, mode=0, blocksize=256, ignore_half_block_zeros=True, de
                 return
         elif mode == 2:  # per segment
             for i, segment in enumerate(binary.segments):
-                _handle("segment #%d" % i, segment.content)
+                _handle("segment #%d" % i, segment.content.tobytes())
         else:
             raise NotImplementedError("This mode does not exist")
         if logger:
@@ -179,11 +218,11 @@ def characteristics(executable, n_samples=N_SAMPLES, window_size=lambda s: 2*s,
     :param n_samples:   number of samples of entropy required
     :param window_size: window size for computing the entropy
     """
-    data = {'name': os.path.basename(executable), 'entropy': [], 'sections': []}
+    data, path = {'name': os.path.basename(executable), 'entropy': [], 'sections': []}, str(executable)
     # compute window-based entropy
-    with open(str(executable), "rb") as f:
+    with open(path, "rb") as f:
         data['entropy*'] = entropy(f.read())
-    with open(str(executable), "rb") as f:
+    with open(path, "rb") as f:
         size = data['size'] = os.fstat(f.fileno()).st_size
         step = abs(size // n_samples)
         chunksize = data['chunksize'] = size / n_samples
@@ -204,31 +243,36 @@ def characteristics(executable, n_samples=N_SAMPLES, window_size=lambda s: 2*s,
     # compute other characteristics using LIEF (catch warnings from stderr)
     tmp_fd, null_fd = os.dup(2), os.open(os.devnull, os.O_RDWR)
     os.dup2(null_fd, 2)
-    binary = lief.parse(str(executable))
+    binary = lief.parse(path)
     os.dup2(tmp_fd, 2)  # restore stderr
     os.close(null_fd)
     if binary is None:
         raise TypeError("Not an executable")
-    data['type'] = "ELF" if type(binary) is lief.ELF.Binary else \
-                   "Mach-O" if type(binary) is lief.MachO.Binary else \
-                   "PE" if type(binary) is lief.PE.Binary else None
-    if data['type'] is None:
-        raise TypeError("Unsupported executable format")
+    data['type'] = __btype(binary)
     # entry point (EP)
-    ep = binary.rva_to_offset(binary.optional_header.addressof_entrypoint)
+    ep, ep_section = _get_ep_and_section(binary)
+    # convert to 3-tuple (EP offset on plot, EP file offset, section name containing EP)
+    data['entrypoint'] = None if ep is None else (int(ep // data['chunksize']), ep, __secname(ep_section))
     # sections
-    __d = lambda s, e, n: (s, e, n, statistics.mean(data['entropy'][s:e+1]))
-    data['sections'] = [__d(0, int(max(MIN_ZONE_WIDTH, binary.sizeof_headers // chunksize)), "Headers")]
-    data['entrypoint'] = int(ep // data['chunksize'])
-    ep_section = binary.section_from_rva(binary.optional_header.addressof_entrypoint).name.rstrip("\x00")
-    # convert to 3-tuple (real EP, EP, section of real EP)
-    data['entrypoint'] = (data['entrypoint'], ep, __secname(ep_section))
-    for section in sorted(binary.sections, key=lambda x: x.offset):
+    __d = lambda s, e, n: (int(s), int(e), n, statistics.mean(data['entropy'][int(s):int(e)+1]))
+    data['sections'] = [__d(0, int(max(MIN_ZONE_WIDTH, binary.sections[0].offset // chunksize)), "Headers")] \
+                       if len(binary.sections) > 0 else []
+    for i, section in enumerate(binary.sections):
         name = __secname(section.name)
+        # special case: for some executables compiled with debug information, sections may be of the form "/[N]" (where
+        #                N is the offset in the string table ; in this case, we compute the real section names)
+        if re.match(r"^\/\d+$", name) and _real_section_names(path):
+            name = _real_section_names(path)[i]
         start = max(data['sections'][-1][1] if len(data['sections']) > 0 else 0, int(section.offset // chunksize))
         max_end = min(max(start + MIN_ZONE_WIDTH, int((section.offset + section.size) // chunksize)),
                       len(data['entropy']) - 1)
         data['sections'].append(__d(int(min(start, max_end - MIN_ZONE_WIDTH)), int(max_end), name))
+    # adjust the entry point (be sure that its position on the plot is within the EP section)
+    if ep:
+        ep_pos, _, ep_sec_name = data['entrypoint']
+        for s, e, name, m in data['sections']:
+            if name == ep_sec_name:
+                data['entrypoint'] = (min(max(ep_pos, s), e), ep, ep_sec_name)
     # fill in undefined sections
     prev_end = None
     for i, t in enumerate(data['sections'][:]):
@@ -236,10 +280,16 @@ def characteristics(executable, n_samples=N_SAMPLES, window_size=lambda s: 2*s,
         if prev_end and prev_end < start:
             data['sections'].insert(i, __d(prev_end, start, "<undef>"))
         prev_end = end
-    # add overlay
-    last = data['sections'][-1][1]
-    if last + 1 < n_samples:
-        data['sections'].append(__d(int(last), int(n_samples), "Overlay"))
+    if len(binary.sections) > 0:
+        last = data['sections'][-1][1]
+        if data['type'] == "ELF":
+            # add section header table
+            sh_size = binary.header.section_header_size * binary.header.numberof_sections
+            data['sections'].append(__d(int(last), int(last) + sh_size // chunksize, "Header"))
+        elif data['type'] == "PE":
+            # add overlay
+            if last + 1 < n_samples:
+                data['sections'].append(__d(int(last), int(n_samples), "Overlay"))
     return data
 
 
@@ -253,7 +303,7 @@ def entropy(something, blocksize=0, ignore_half_block_zeros=False):
     """
     e, l = [], len(something)
     if l == 0:
-        return 0.
+        return ([], None) if blocksize > 0 else 0.
     bs = blocksize or l
     n_blocks, n_ignored = math.ceil(float(l) / bs), 0
     for i in range(0, l, bs):
@@ -275,13 +325,10 @@ def entropy(something, blocksize=0, ignore_half_block_zeros=False):
             n_ignored += 1
             continue
         # if not ignored, process it
-        d = {}
-        for c in block:
-            d.setdefault(c, 0)
-            d[c] += 1
-        e.append(-sum([p * math.log2(p) for p in [float(ctr) / lb for ctr in d.values()]]) or .0)
+        chr_cts = [block.count(c) for c in set(block)]
+        e.append(-sum([p * math.log2(p) for p in [float(ctr) / lb for ctr in chr_cts]]) or .0)
     # return the entropies per block and the average entropy of all blocks if n_blocks > 1
-    return (e, sum([n or 0 for n in e]) / ((n_blocks - n_ignored) or 1)) if n_blocks > 1 else e[0]
+    return (e, sum(n or 0 for n in e) / ((n_blocks - n_ignored) or 1)) if n_blocks > 1 or blocksize > 0 else e[0]
 
 
 def is_packed(entropies, average, threshold_average_entropy, threshold_highest_entropy, logger=None):
@@ -309,7 +356,8 @@ def is_packed(entropies, average, threshold_average_entropy, threshold_highest_e
     return c1 and c2
 
 
-def plot(*filenames, img_name=None, img_format="png", dpi=200, labels=None, sublabel=None, scale=False, **kwargs):
+def plot(*filenames, img_name=None, img_format="png", dpi=200, labels=None, sublabel=None, scale=False, target=None,
+         return_plt=False, **kwargs):
     """ Plot input binaries' characteristics on a same figure.
     
     :param img:        destination filename for the created figure
@@ -318,8 +366,14 @@ def plot(*filenames, img_name=None, img_format="png", dpi=200, labels=None, subl
     :param dpi:        resolution of the created figure
     :param labels:     list of custom labels to be used for the binaries (can be lambda-based)
     :param sublabel:   static or lambda-based sublabel for display under the label
+    :param scale:      use the same scale for multiple files
+    :param target:     use this target name in the title (instead of the first filename from the input list)
+    :param return_plt: return plt object (if False, close the current figure and return None)
     :param kwargs:     keyword-arguments for characteristics(...) ; n_samples and window_size
     """
+    import matplotlib.pyplot as plt
+    from matplotlib.patches import Patch
+    plt.rcParams['font.family'] = "serif"
     if len(filenames) == 0:
         raise ValueError("No executable to plot")
     lloc, title = kwargs.get('legend_location', "lower center"), not kwargs.get('no_title', False)
@@ -343,8 +397,9 @@ def plot(*filenames, img_name=None, img_format="png", dpi=200, labels=None, subl
         obj.axis("off")
         # set the main title for the whole figure
         if i == 0 and title:
-            fig.suptitle("Entropy per section of %s file: %s" % (data['type'], filename), x=[.6, .5][labels is None],
-                         y=1.-.6/(nf+[0, 1][title]), ha="center", va="bottom", fontsize="xx-large", fontweight="bold")
+            x_t, y_t = [.6, .5][labels is None], 1. - .6 / (nf + [0, 1][title])
+            fig.suptitle("Entropy per section of %s file: %s" % (data['type'], target or filename), x=x_t, y=y_t,
+                         ha="center", va="bottom", fontsize="xx-large", fontweight="bold")
         # set the label and sublabel and display them
         try:
             label = labels[i]
@@ -353,7 +408,7 @@ def plot(*filenames, img_name=None, img_format="png", dpi=200, labels=None, subl
         except:
             pass
         ref_point = .65
-        if sublabel:
+        if sublabel and not (isinstance(sublabel, str) and "ep" in sublabel and data['entrypoint'] is None):
             if isinstance(sublabel, str):
                 sublabel = SUBLABELS.get(sublabel)
             sl = sublabel(data) if isinstance(sublabel, type(lambda: 0)) else None
@@ -372,9 +427,10 @@ def plot(*filenames, img_name=None, img_format="png", dpi=200, labels=None, subl
                 y_pos = min(1., ref_point + nl * [.16, .12, .09, .08][min(4, nl)-1])
             obj.text(s=label, x=-420., y=y_pos, fontsize="large", ha="left", va="center")
         # display the entry point
-        obj.vlines(x=data['entrypoint'][0], ymin=0, ymax=1, color="r", zorder=11).set_label("Entry point")
-        obj.text(data['entrypoint'][0], -.15, "______", c="r", ha="center", rotation=90, size=.8,
-                 bbox={'boxstyle': "rarrow", 'fc': "r", 'ec': "r", 'lw': 1})
+        if data['entrypoint']:
+            obj.vlines(x=data['entrypoint'][0], ymin=0, ymax=1, color="r", zorder=11).set_label("Entry point")
+            obj.text(data['entrypoint'][0], -.15, "______", c="r", ha="center", rotation=90, size=.8,
+                     bbox={'boxstyle': "rarrow", 'fc': "r", 'ec': "r", 'lw': 1})
         color_cursor, last = 0, None
         for start, end, name, avg_ent in data['sections']:
             x = range(start, min(n, end + 1))
@@ -404,11 +460,17 @@ def plot(*filenames, img_name=None, img_format="png", dpi=200, labels=None, subl
             obj.fill_between(x, [0] * len(x), data['entropy'][start:end+1], facecolor=c)
             l = obj.hlines(y=statistics.mean(data['entropy'][start:end+1]), xmin=x[0], xmax=x[-1], color="black",
                            linestyle=(0, (5, 5)), linewidth=.5)
-        l.set_label("Average entropy of section")
+        if len(data['sections']) > 0:
+            l.set_label("Average entropy of section")
+        else:
+            obj.text(.5, ref_point, "Could not parse sections", fontsize=16, color="red", ha="center", va="center")
     plt.subplots_adjust(left=[.15, .02][labels is None and sublabel is None], right=[1.02, .82][lloc_side],
                         bottom=.5/max(1.75, nf))
     h, l = (objs[[0, 1][title]] if nf+[0, 1][title] > 1 else objs).get_legend_handles_labels()
-    plt.figlegend(h, l, loc=lloc, ncol=1 if lloc_side else 2, prop={'size': 7})
+    h.append(Patch(facecolor="black")), l.append("Headers")
+    h.append(Patch(facecolor="lightgray")), l.append("Overlay")
+    if len(h) > 0:
+        plt.figlegend(h, l, loc=lloc, ncol=1 if lloc_side else len(l), prop={'size': 7})
     img_name = img_name or os.path.splitext(os.path.basename(filenames[0]))[0]
     # appending the extension to img_name is necessary for avoiding an error when the filename contains a ".[...]" ;
     #  e.g. "PortableWinCDEmu-4.0" => this fails with "ValueError: Format '0' is not supported"
@@ -416,5 +478,7 @@ def plot(*filenames, img_name=None, img_format="png", dpi=200, labels=None, subl
         plt.savefig(img_name + "." + img_format, img_format=img_format, dpi=dpi, bbox_inches="tight")
     except:  # format argument renamed in further versions of pyplot
         plt.savefig(img_name + "." + img_format, format=img_format, dpi=dpi, bbox_inches="tight")
-    return plt
+    if return_plt:
+        return plt
+    plt.close()
 
